"""Home automation cluster handlers module for Zigbee Home Automation."""

from __future__ import annotations

import enum

from zigpy.zcl.clusters.homeautomation import (
    ApplianceEventAlerts,
    ApplianceIdentification,
    ApplianceStatistics,
    Diagnostic,
    ElectricalMeasurement,
    MeterIdentification,
)

from zha.zigbee.cluster_handlers import AttrReportConfig, ClusterHandler, registries
from zha.zigbee.cluster_handlers.const import (
    CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT,
    REPORT_CONFIG_IMMEDIATE,
    REPORT_CONFIG_OP,
)


@registries.CLUSTER_HANDLER_REGISTRY.register(ApplianceEventAlerts.cluster_id)
class ApplianceEventAlertsClusterHandler(ClusterHandler):
    """Appliance Event Alerts cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(ApplianceIdentification.cluster_id)
class ApplianceIdentificationClusterHandler(ClusterHandler):
    """Appliance Identification cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(ApplianceStatistics.cluster_id)
class ApplianceStatisticsClusterHandler(ClusterHandler):
    """Appliance Statistics cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(Diagnostic.cluster_id)
class DiagnosticClusterHandler(ClusterHandler):
    """Diagnostic cluster handler."""


@registries.CLUSTER_HANDLER_REGISTRY.register(ElectricalMeasurement.cluster_id)
class ElectricalMeasurementClusterHandler(ClusterHandler):
    """Cluster handler that polls active power level."""

    CLUSTER_HANDLER_NAME = CLUSTER_HANDLER_ELECTRICAL_MEASUREMENT

    class MeasurementType(enum.IntFlag):
        """Measurement types."""

        ACTIVE_MEASUREMENT = 1
        REACTIVE_MEASUREMENT = 2
        APPARENT_MEASUREMENT = 4
        PHASE_A_MEASUREMENT = 8
        PHASE_B_MEASUREMENT = 16
        PHASE_C_MEASUREMENT = 32
        DC_MEASUREMENT = 64
        HARMONICS_MEASUREMENT = 128
        POWER_QUALITY_MEASUREMENT = 256

    REPORT_CONFIG = (
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_voltage_multiplier.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_voltage_divisor.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_current_multiplier.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_current_divisor.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_power_multiplier.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_power_divisor.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.power_multiplier.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.power_divisor.name,
            config=REPORT_CONFIG_IMMEDIATE,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.active_power.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.active_power_ph_b.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.active_power_ph_c.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.total_active_power.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.apparent_power.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_current.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_current_ph_b.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_current_ph_c.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_voltage.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_voltage_ph_b.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.rms_voltage_ph_c.name,
            config=REPORT_CONFIG_OP,
        ),
        AttrReportConfig(
            attr=ElectricalMeasurement.AttributeDefs.ac_frequency.name,
            config=REPORT_CONFIG_OP,
        ),
    )
    ZCL_POLLING_ATTRS = [
        ElectricalMeasurement.AttributeDefs.ac_frequency.name,
        ElectricalMeasurement.AttributeDefs.ac_frequency_max.name,
        ElectricalMeasurement.AttributeDefs.active_power.name,
        ElectricalMeasurement.AttributeDefs.active_power_ph_b.name,
        ElectricalMeasurement.AttributeDefs.active_power_ph_c.name,
        ElectricalMeasurement.AttributeDefs.active_power_max.name,
        ElectricalMeasurement.AttributeDefs.active_power_max_ph_b.name,
        ElectricalMeasurement.AttributeDefs.active_power_max_ph_c.name,
        ElectricalMeasurement.AttributeDefs.total_active_power.name,
        ElectricalMeasurement.AttributeDefs.apparent_power.name,
        ElectricalMeasurement.AttributeDefs.power_factor.name,
        ElectricalMeasurement.AttributeDefs.power_factor_ph_b.name,
        ElectricalMeasurement.AttributeDefs.power_factor_ph_c.name,
        ElectricalMeasurement.AttributeDefs.rms_current.name,
        ElectricalMeasurement.AttributeDefs.rms_current_ph_b.name,
        ElectricalMeasurement.AttributeDefs.rms_current_ph_c.name,
        ElectricalMeasurement.AttributeDefs.rms_current_max.name,
        ElectricalMeasurement.AttributeDefs.rms_current_max_ph_b.name,
        ElectricalMeasurement.AttributeDefs.rms_current_max_ph_c.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage_ph_b.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage_ph_c.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max_ph_b.name,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max_ph_c.name,
    ]
    ZCL_INIT_ATTRS = {
        ElectricalMeasurement.AttributeDefs.ac_frequency_divisor.name: True,
        ElectricalMeasurement.AttributeDefs.ac_frequency_max.name: True,
        ElectricalMeasurement.AttributeDefs.ac_frequency_multiplier.name: True,
        ElectricalMeasurement.AttributeDefs.active_power_max.name: True,
        ElectricalMeasurement.AttributeDefs.active_power_max_ph_b.name: True,
        ElectricalMeasurement.AttributeDefs.active_power_max_ph_c.name: True,
        ElectricalMeasurement.AttributeDefs.measurement_type.name: True,
        ElectricalMeasurement.AttributeDefs.power_factor.name: True,
        ElectricalMeasurement.AttributeDefs.power_factor_ph_b.name: True,
        ElectricalMeasurement.AttributeDefs.power_factor_ph_c.name: True,
        ElectricalMeasurement.AttributeDefs.rms_current_max.name: True,
        ElectricalMeasurement.AttributeDefs.rms_current_max_ph_b.name: True,
        ElectricalMeasurement.AttributeDefs.rms_current_max_ph_c.name: True,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max.name: True,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max_ph_b.name: True,
        ElectricalMeasurement.AttributeDefs.rms_voltage_max_ph_c.name: True,
    }

    async def async_update(self):
        """Retrieve latest state."""
        self.debug("async_update")

        # This is a polling cluster handler. Don't allow cache.
        attrs = [
            attr
            for attr in self.ZCL_POLLING_ATTRS
            if attr not in self.cluster.unsupported_attributes
        ]
        await self.get_attributes(attrs, from_cache=False, only_cache=False)

    @property
    def ac_current_divisor(self) -> int:
        """Return ac current divisor."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_current_divisor.name
            )
            or 1
        )

    @property
    def ac_current_multiplier(self) -> int:
        """Return ac current multiplier."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_current_multiplier.name
            )
            or 1
        )

    @property
    def ac_voltage_divisor(self) -> int:
        """Return ac voltage divisor."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_voltage_divisor.name
            )
            or 1
        )

    @property
    def ac_voltage_multiplier(self) -> int:
        """Return ac voltage multiplier."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_voltage_multiplier.name
            )
            or 1
        )

    @property
    def ac_frequency_divisor(self) -> int:
        """Return ac frequency divisor."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_frequency_divisor.name
            )
            or 1
        )

    @property
    def ac_frequency_multiplier(self) -> int:
        """Return ac frequency multiplier."""
        return (
            self.cluster.get(
                ElectricalMeasurement.AttributeDefs.ac_frequency_multiplier.name
            )
            or 1
        )

    @property
    def ac_power_divisor(self) -> int:
        """Return active power divisor."""
        return self.cluster.get(
            ElectricalMeasurement.AttributeDefs.ac_power_divisor.name,
            self.cluster.get(ElectricalMeasurement.AttributeDefs.power_divisor.name)
            or 1,
        )

    @property
    def ac_power_multiplier(self) -> int:
        """Return active power divisor."""
        return self.cluster.get(
            ElectricalMeasurement.AttributeDefs.ac_power_multiplier.name,
            self.cluster.get(ElectricalMeasurement.AttributeDefs.power_multiplier.name)
            or 1,
        )

    @property
    def measurement_type(self) -> str | None:
        """Return Measurement type."""
        if (
            meas_type := self.cluster.get(
                ElectricalMeasurement.AttributeDefs.measurement_type.name
            )
        ) is None:
            return None

        meas_type = self.MeasurementType(meas_type)
        return ", ".join(
            m.name
            for m in self.MeasurementType
            if m in meas_type and m.name is not None
        )


@registries.CLUSTER_HANDLER_REGISTRY.register(MeterIdentification.cluster_id)
class MeterIdentificationClusterHandler(ClusterHandler):
    """Metering Identification cluster handler."""



"""Debounce helper."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from logging import Logger
from typing import TYPE_CHECKING, Generic, TypeVar

from zha.async_ import ZHAJob

if TYPE_CHECKING:
    from zha.application.gateway import Gateway


_R_co = TypeVar("_R_co", covariant=True)


class Debouncer(Generic[_R_co]):  # noqa: UP046
    """Class to rate limit calls to a specific command."""

    def __init__(
        self,
        gateway: Gateway,
        logger: Logger,
        *,
        cooldown: float,
        immediate: bool,
        function: Callable[[], _R_co] | None = None,
        background: bool = False,
    ) -> None:
        """Initialize debounce.

        immediate: indicate if the function needs to be called right away and
                   wait <cooldown> until executing next invocation.
        function: optional and can be instantiated later.
        """
        self.gateway = gateway
        self.logger = logger
        self._function = function
        self.cooldown = cooldown
        self.immediate = immediate
        self._timer_task: asyncio.TimerHandle | None = None
        self._execute_at_end_of_timer: bool = False
        self._execute_lock = asyncio.Lock()
        self._background = background
        self._job: ZHAJob[[], _R_co] | None = (
            None
            if function is None
            else ZHAJob(
                function, f"debouncer cooldown={cooldown}, immediate={immediate}"
            )
        )
        self._shutdown_requested = False

    @property
    def function(self) -> Callable[[], _R_co] | None:
        """Return the function being wrapped by the Debouncer."""
        return self._function

    @function.setter
    def function(self, function: Callable[[], _R_co]) -> None:
        """Update the function being wrapped by the Debouncer."""
        self._function = function
        if self._job is None or function != self._job.target:
            self._job = ZHAJob(
                function,
                f"debouncer cooldown={self.cooldown}, immediate={self.immediate}",
            )

    def async_schedule_call(self) -> None:
        """Schedule a call to the function."""
        if self._async_schedule_or_call_now():
            self._execute_at_end_of_timer = True
            self._on_debounce()

    def _async_schedule_or_call_now(self) -> bool:
        """Check if a call should be scheduled.

        Returns True if the function should be called immediately.

        Returns False if there is nothing to do.
        """
        if self._shutdown_requested:
            self.logger.debug("Debouncer call ignored as shutdown has been requested.")
            return False

        if self._timer_task:
            if not self._execute_at_end_of_timer:
                self._execute_at_end_of_timer = True

            return False

        # Locked means a call is in progress. Any call is good, so abort.
        if self._execute_lock.locked():
            return False

        if not self.immediate:
            self._execute_at_end_of_timer = True
            self._schedule_timer()
            return False

        return True

    async def async_call(self) -> None:
        """Call the function."""
        if not self._async_schedule_or_call_now():
            return

        async with self._execute_lock:
            # Abort if timer got set while we're waiting for the lock.
            if self._timer_task:
                return

            assert self._job is not None
            try:
                if task := self.gateway.async_run_zha_job(
                    self._job, background=self._background
                ):
                    await task
            finally:
                self._schedule_timer()

    async def _handle_timer_finish(self) -> None:
        """Handle a finished timer."""
        assert self._job is not None

        self._execute_at_end_of_timer = False

        # Locked means a call is in progress. Any call is good, so abort.
        if self._execute_lock.locked():
            return

        async with self._execute_lock:
            # Abort if timer got set while we're waiting for the lock.
            if self._timer_task:
                return

            try:
                if task := self.gateway.async_run_zha_job(
                    self._job, background=self._background
                ):
                    await task
            except Exception:  # pylint: disable=broad-except
                self.logger.exception("Unexpected exception from %s", self.function)
            finally:
                # Schedule a new timer to prevent new runs during cooldown
                self._schedule_timer()

    def async_shutdown(self) -> None:
        """Cancel any scheduled call, and prevent new runs."""
        self._shutdown_requested = True
        self.async_cancel()

    def async_cancel(self) -> None:
        """Cancel any scheduled call."""
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

        self._execute_at_end_of_timer = False

    def _on_debounce(self) -> None:
        """Create job task, but only if pending."""
        self._timer_task = None
        if not self._execute_at_end_of_timer:
            return
        self._execute_at_end_of_timer = False
        name = f"debouncer {self._job} finish cooldown={self.cooldown}, immediate={self.immediate}"
        if not self._background:
            self.gateway.async_create_task(
                self._handle_timer_finish(), name, eager_start=True
            )
            return
        self.gateway.async_create_background_task(
            self._handle_timer_finish(), name, eager_start=True
        )

    def _schedule_timer(self) -> None:
        """Schedule a timer."""
        if not self._shutdown_requested:
            self._timer_task = self.gateway.loop.call_later(
                self.cooldown, self._on_debounce
            )



"""Constants for ZHA."""

from __future__ import annotations

import enum
from typing import Final

import bellows.zigbee.application
import zigpy.application
import zigpy.types as t
import zigpy_deconz.zigbee.application
import zigpy_xbee.zigbee.application
import zigpy_zigate.zigbee.application
import zigpy_znp.zigbee.application

ATTR_ACTIVE_COORDINATOR = "active_coordinator"
ATTR_ARGS = "args"
ATTR_ATTRIBUTE = "attribute"
ATTR_ATTRIBUTE_ID = "attribute_id"
ATTR_ATTRIBUTE_NAME = "attribute_name"
ATTR_AVAILABLE = "available"
ATTR_CLUSTER_ID = "cluster_id"
ATTR_CLUSTER_TYPE = "cluster_type"
ATTR_COMMAND: Final = "command"
ATTR_COMMAND_TYPE = "command_type"
ATTR_DEVICE_IEEE = "device_ieee"
ATTR_DEVICE_TYPE = "device_type"
ATTR_ENDPOINTS = "endpoints"
ATTR_ENDPOINT_NAMES = "endpoint_names"
ATTR_ENDPOINT_ID = "endpoint_id"
ATTR_IEEE = "ieee"
ATTR_IN_CLUSTERS = "in_clusters"
ATTR_LAST_SEEN = "last_seen"
ATTR_LEVEL = "level"
ATTR_LQI = "lqi"
ATTR_MANUFACTURER = "manufacturer"
ATTR_MANUFACTURER_CODE = "manufacturer_code"
ATTR_MEMBERS = "members"
ATTR_MODEL = "model"
ATTR_NAME = "name"
ATTR_NEIGHBORS = "neighbors"
ATTR_NODE_DESCRIPTOR = "node_descriptor"
ATTR_NWK = "nwk"
ATTR_OUT_CLUSTERS = "out_clusters"
ATTR_PARAMS = "params"
ATTR_POWER_SOURCE = "power_source"
ATTR_PROFILE_ID = "profile_id"
ATTR_QUIRK_APPLIED = "quirk_applied"
ATTR_QUIRK_CLASS = "quirk_class"
ATTR_QUIRK_ID = "quirk_id"
ATTR_ROUTES = "routes"
ATTR_RSSI = "rssi"
ATTR_SIGNATURE = "signature"
ATTR_TYPE = "type"
ATTR_UNIQUE_ID = "unique_id"
ATTR_VALUE = "value"
ATTR_WARNING_DEVICE_DURATION = "duration"
ATTR_WARNING_DEVICE_MODE = "mode"
ATTR_WARNING_DEVICE_STROBE = "strobe"
ATTR_WARNING_DEVICE_STROBE_DUTY_CYCLE = "duty_cycle"
ATTR_WARNING_DEVICE_STROBE_INTENSITY = "intensity"

# Class of device within its domain
ATTR_DEVICE_CLASS: Final = "device_class"

BAUD_RATES = [2400, 4800, 9600, 14400, 19200, 38400, 57600, 115200, 128000, 256000]

CLUSTER_DETAILS = "cluster_details"

CLUSTER_COMMAND_SERVER = "server"
CLUSTER_COMMANDS_CLIENT = "client_commands"
CLUSTER_COMMANDS_SERVER = "server_commands"
CLUSTER_TYPE_IN = "in"
CLUSTER_TYPE_OUT = "out"

CONF_USE_THREAD = "use_thread"

CONF_DEFAULT_CONSIDER_UNAVAILABLE_MAINS = 60 * 60 * 2  # 2 hours
CONF_DEFAULT_CONSIDER_UNAVAILABLE_BATTERY = 60 * 60 * 6  # 6 hours

POWER_MAINS_POWERED = "Mains"
POWER_BATTERY_OR_UNKNOWN = "Battery or Unknown"

# Device is in away mode
PRESET_AWAY = "away"
PRESET_SCHEDULE = "Schedule"
PRESET_COMPLEX = "Complex"
PRESET_TEMP_MANUAL = "Temporary manual"
# Device turn all valve full up
PRESET_BOOST = "boost"
# No preset is active
PRESET_NONE = "none"

ENTITY_METADATA = "entity_metadata"

ZCL_INIT_ATTRS = "ZCL_INIT_ATTRS"
REPORT_CONFIG = "REPORT_CONFIG"

_ControllerClsType = type[zigpy.application.ControllerApplication]


class RadioType(enum.Enum):
    """Possible options for radio type."""

    ezsp = (
        "EZSP = Silicon Labs EmberZNet protocol: Elelabs, HUSBZB-1, Telegesis",
        bellows.zigbee.application.ControllerApplication,
    )
    znp = (
        "ZNP = Texas Instruments Z-Stack ZNP protocol: CC253x, CC26x2, CC13x2",
        zigpy_znp.zigbee.application.ControllerApplication,
    )
    deconz = (
        "deCONZ = dresden elektronik deCONZ protocol: ConBee I/II, RaspBee I/II",
        zigpy_deconz.zigbee.application.ControllerApplication,
    )
    zigate = (
        "ZiGate = ZiGate Zigbee radios: PiZiGate, ZiGate USB-TTL, ZiGate WiFi",
        zigpy_zigate.zigbee.application.ControllerApplication,
    )
    xbee = (
        "XBee = Digi XBee Zigbee radios: Digi XBee Series 2, 2C, 3",
        zigpy_xbee.zigbee.application.ControllerApplication,
    )

    @classmethod
    def list(cls) -> list[str]:
        """Return a list of descriptions."""
        return [e.description for e in RadioType]

    @classmethod
    def get_by_description(cls, description: str) -> RadioType:
        """Get radio by description."""
        for radio in cls:
            if radio.description == description:
                return radio
        raise ValueError

    def __init__(self, description: str, controller_cls: _ControllerClsType) -> None:
        """Init instance."""
        self._desc = description
        self._ctrl_cls = controller_cls

    @property
    def controller(self) -> _ControllerClsType:
        """Return controller class."""
        return self._ctrl_cls

    @property
    def description(self) -> str:
        """Return radio type description."""
        return self._desc

    @property
    def pretty_name(self) -> str:
        """Return radio type name."""
        return self.description.split(" = ", 1)[0]


UNKNOWN = "unknown"
UNKNOWN_MANUFACTURER = "unk_manufacturer"
UNKNOWN_MODEL = "unk_model"

WARNING_DEVICE_MODE_STOP = 0
WARNING_DEVICE_MODE_BURGLAR = 1
WARNING_DEVICE_MODE_FIRE = 2
WARNING_DEVICE_MODE_EMERGENCY = 3
WARNING_DEVICE_MODE_POLICE_PANIC = 4
WARNING_DEVICE_MODE_FIRE_PANIC = 5
WARNING_DEVICE_MODE_EMERGENCY_PANIC = 6

WARNING_DEVICE_STROBE_NO = 0
WARNING_DEVICE_STROBE_YES = 1

WARNING_DEVICE_SOUND_LOW = 0
WARNING_DEVICE_SOUND_MEDIUM = 1
WARNING_DEVICE_SOUND_HIGH = 2
WARNING_DEVICE_SOUND_VERY_HIGH = 3

WARNING_DEVICE_STROBE_LOW = 0x00
WARNING_DEVICE_STROBE_MEDIUM = 0x01
WARNING_DEVICE_STROBE_HIGH = 0x02
WARNING_DEVICE_STROBE_VERY_HIGH = 0x03

WARNING_DEVICE_SQUAWK_MODE_ARMED = 0
WARNING_DEVICE_SQUAWK_MODE_DISARMED = 1

ZHA_CLUSTER_HANDLER_MSG = "zha_channel_message"
ZHA_CLUSTER_HANDLER_MSG_BIND = "zha_channel_bind"
ZHA_CLUSTER_HANDLER_MSG_CFG_RPT = "zha_channel_configure_reporting"
ZHA_CLUSTER_HANDLER_MSG_DATA = "zha_channel_msg_data"
ZHA_CLUSTER_HANDLER_CFG_DONE = "zha_channel_cfg_done"
ZHA_CLUSTER_HANDLER_READS_PER_REQ = 5
ZHA_EVENT = "zha_event"
ZHA_DEVICE_UPDATED_EVENT = "zha_device_updated_event"
ZHA_GW_MSG = "zha_gateway_message"
ZHA_GW_MSG_DEVICE_FULL_INIT = "device_fully_initialized"
ZHA_GW_MSG_DEVICE_INFO = "device_info"
ZHA_GW_MSG_DEVICE_JOINED = "device_joined"
ZHA_GW_MSG_DEVICE_LEFT = "device_left"
ZHA_GW_MSG_DEVICE_REMOVED = "device_removed"
ZHA_GW_MSG_GROUP_ADDED = "group_added"
ZHA_GW_MSG_GROUP_INFO = "group_info"
ZHA_GW_MSG_GROUP_MEMBER_ADDED = "group_member_added"
ZHA_GW_MSG_GROUP_MEMBER_REMOVED = "group_member_removed"
ZHA_GW_MSG_GROUP_REMOVED = "group_removed"
ZHA_GW_MSG_LOG_ENTRY = "log_entry"
ZHA_GW_MSG_LOG_OUTPUT = "log_output"
ZHA_GW_MSG_RAW_INIT = "raw_device_initialized"
ZHA_GW_MSG_CONNECTION_LOST = "connection_lost"


class Strobe(t.enum8):
    """Strobe enum."""

    No_Strobe = 0x00
    Strobe = 0x01


class UniqueIdMigration(enum.Enum):
    """Migrateion root for unique ID."""

    LEGACY_CLUSTER = "legacy_cluster"
    LEGACY_ENDPOINT = "legacy_endpoint"

    DEVICE = "device"
    CLUSTER = "cluster"
    ENDPOINT = "endpoint"



"""Helpers for Zigbee Home Automation."""

from __future__ import annotations

import ast
import asyncio
import binascii
import collections
from collections.abc import Callable
import contextlib
import dataclasses
from dataclasses import dataclass
import datetime
import enum
import logging
import re
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar

import voluptuous as vol
import zigpy.exceptions
import zigpy.types
import zigpy.util
import zigpy.zcl
from zigpy.zcl.foundation import CommandSchema
import zigpy.zdo.types as zdo_types

from zha.application import Platform
from zha.application.const import (
    CLUSTER_TYPE_IN,
    CLUSTER_TYPE_OUT,
    CONF_DEFAULT_CONSIDER_UNAVAILABLE_BATTERY,
    CONF_DEFAULT_CONSIDER_UNAVAILABLE_MAINS,
)
from zha.async_ import gather_with_limited_concurrency
from zha.decorators import periodic

if TYPE_CHECKING:
    from zha.application.gateway import Gateway
    from zha.zigbee.cluster_handlers import ClusterHandler
    from zha.zigbee.device import Device

_ClusterHandlerT = TypeVar("_ClusterHandlerT", bound="ClusterHandler")
_T = TypeVar("_T")
_R = TypeVar("_R")
_P = ParamSpec("_P")
_LOGGER = logging.getLogger(__name__)


@dataclass
class BindingPair:
    """Information for binding."""

    source_cluster: zigpy.zcl.Cluster
    target_ieee: zigpy.types.EUI64
    target_ep_id: int

    @property
    def destination_address(self) -> zdo_types.MultiAddress:
        """Return a ZDO multi address instance."""
        return zdo_types.MultiAddress(
            addrmode=3, ieee=self.target_ieee, endpoint=self.target_ep_id
        )


async def safe_read(
    cluster: zigpy.zcl.Cluster,
    attributes: list[int | str],
    allow_cache: bool = True,
    only_cache: bool = False,
    manufacturer=None,
):
    """Swallow all exceptions from network read.

    If we throw during initialization, setup fails. Rather have an entity that
    exists, but is in a maybe wrong state, than no entity. This method should
    probably only be used during initialization.
    """
    try:
        result, _ = await cluster.read_attributes(
            attributes,
            allow_cache=allow_cache,
            only_cache=only_cache,
            manufacturer=manufacturer,
        )
        return result
    except Exception:  # pylint: disable=broad-except
        return {}


async def get_matched_clusters(
    source_zha_device: Device, target_zha_device: Device
) -> list[BindingPair]:
    """Get matched input/output cluster pairs for 2 devices."""
    from zha.zigbee.cluster_handlers.registries import (  # pylint: disable=import-outside-toplevel  # noqa: PLC0415
        BINDABLE_CLUSTERS,
    )

    source_clusters = source_zha_device.async_get_std_clusters()
    target_clusters = target_zha_device.async_get_std_clusters()
    clusters_to_bind = []

    for endpoint_id in source_clusters:
        for cluster_id in source_clusters[endpoint_id][CLUSTER_TYPE_OUT]:
            if cluster_id not in BINDABLE_CLUSTERS:
                continue
            if target_zha_device.nwk == 0x0000:
                cluster_pair = BindingPair(
                    source_cluster=source_clusters[endpoint_id][CLUSTER_TYPE_OUT][
                        cluster_id
                    ],
                    target_ieee=target_zha_device.ieee,
                    target_ep_id=target_zha_device.device.application.get_endpoint_id(
                        cluster_id, is_server_cluster=True
                    ),
                )
                clusters_to_bind.append(cluster_pair)
                continue
            for t_endpoint_id in target_clusters:
                if cluster_id in target_clusters[t_endpoint_id][CLUSTER_TYPE_IN]:
                    cluster_pair = BindingPair(
                        source_cluster=source_clusters[endpoint_id][CLUSTER_TYPE_OUT][
                            cluster_id
                        ],
                        target_ieee=target_zha_device.ieee,
                        target_ep_id=t_endpoint_id,
                    )
                    clusters_to_bind.append(cluster_pair)
    return clusters_to_bind


def convert_zcl_value(value: Any, field_type: Any) -> Any:
    """Convert user input to ZCL value."""
    if issubclass(field_type, enum.Flag):
        if isinstance(value, str):
            with contextlib.suppress(ValueError):
                value = int(value)

        if isinstance(value, int):
            value = field_type(value)
        elif isinstance(value, str):
            # List of flags: `SomeFlag.field1 | field2`
            value = [v.strip() for v in value.split(".", 1)[-1].split("|")]

        if isinstance(value, list):
            new_value = 0

            for flag in value:
                if isinstance(flag, str):
                    new_value |= field_type[flag.replace(" ", "_")]
                else:
                    new_value |= flag

            value = field_type(new_value)
    elif issubclass(field_type, enum.Enum):
        value = (
            field_type[value.replace(" ", "_").split(".", 1)[-1]]
            if isinstance(value, str)
            else field_type(value)
        )
    elif issubclass(field_type, zigpy.types.SerializableBytes):
        if value.startswith(("b'", 'b"')):
            value = ast.literal_eval(value)
        else:
            value = bytes.fromhex(value)

        value = field_type(value)
    elif issubclass(field_type, int):
        if isinstance(value, str) and value.startswith("0x"):
            value = int(value, 16)

        value = field_type(value)
    else:
        value = field_type(value)

    return value


def convert_to_zcl_values(
    fields: dict[str, Any], schema: CommandSchema
) -> dict[str, Any]:
    """Convert user input to ZCL values."""
    converted_fields: dict[str, Any] = {}
    for field in schema.fields:
        if field.name not in fields:
            continue

        value = fields[field.name]
        new_value = converted_fields[field.name] = convert_zcl_value(value, field.type)

        _LOGGER.debug(
            "Converted ZCL schema field(%s) value from: %s to: %s",
            field.name,
            value,
            new_value,
        )

    return converted_fields


def async_is_bindable_target(source_zha_device: Device, target_zha_device: Device):
    """Determine if target is bindable to source."""
    from zha.zigbee.cluster_handlers.registries import (  # pylint: disable=import-outside-toplevel  # noqa: PLC0415
        BINDABLE_CLUSTERS,
    )

    if target_zha_device.nwk == 0x0000:
        return True

    source_clusters = source_zha_device.async_get_std_clusters()
    target_clusters = target_zha_device.async_get_std_clusters()

    for endpoint_id in source_clusters:
        for t_endpoint_id in target_clusters:
            matches = set(
                source_clusters[endpoint_id][CLUSTER_TYPE_OUT].keys()
            ).intersection(target_clusters[t_endpoint_id][CLUSTER_TYPE_IN].keys())
            if any(bindable in BINDABLE_CLUSTERS for bindable in matches):
                return True
    return False


def convert_install_code(value: str) -> zigpy.types.KeyData:
    """Convert string to install code bytes and validate length."""

    try:
        code = binascii.unhexlify(value.replace("-", "").lower())
    except binascii.Error as exc:
        raise vol.Invalid(f"invalid hex string: {value}") from exc

    if len(code) != 18:  # 16 byte code + 2 crc bytes
        raise vol.Invalid("invalid length of the install code")

    link_key = zigpy.util.convert_install_code(code)
    if link_key is None:
        raise vol.Invalid("invalid install code")

    return link_key


QR_CODES = (
    # Consciot
    r"^([\da-fA-F]{16})\|([\da-fA-F]{36})$",
    # Enbrighten
    r"""
        ^Z:
        ([0-9a-fA-F]{16})  # IEEE address
        \$I:
        ([0-9a-fA-F]{36})  # install code
        $
    """,
    # Aqara
    r"""
        \$A:
        ([0-9a-fA-F]{16})  # IEEE address
        \$I:
        ([0-9a-fA-F]{36})  # install code
        $
    """,
    # Bosch
    r"""
        ^RB01SG
        [0-9a-fA-F]{34}
        ([0-9a-fA-F]{16}) # IEEE address
        DLK
        ([0-9a-fA-F]{36}|[0-9a-fA-F]{32}) # install code / link key
        $
    """,
)


def qr_to_install_code(qr_code: str) -> tuple[zigpy.types.EUI64, zigpy.types.KeyData]:
    """Try to parse the QR code.

    if successful, return a tuple of a EUI64 address and install code.
    """

    for code_pattern in QR_CODES:
        match = re.search(code_pattern, qr_code, re.VERBOSE)
        if match is None:
            continue

        ieee_hex = binascii.unhexlify(match[1])
        ieee = zigpy.types.EUI64(ieee_hex[::-1])

        # Bosch supplies (A) device specific link key (DSLK) or (A) install code + crc
        if "RB01SG" in code_pattern and len(match[2]) == 32:
            link_key_hex = binascii.unhexlify(match[2])
            link_key = zigpy.types.KeyData(link_key_hex)
            return ieee, link_key
        install_code = match[2]
        # install_code sanity check
        link_key = convert_install_code(install_code)
        return ieee, link_key

    raise vol.Invalid(f"couldn't convert qr code: {qr_code}")


@dataclass(kw_only=True, slots=True)
class LightOptions:
    """ZHA light options."""

    default_light_transition: float = dataclasses.field(default=0)
    enable_enhanced_light_transition: bool = dataclasses.field(default=False)
    enable_light_transitioning_flag: bool = dataclasses.field(default=True)
    always_prefer_xy_color_mode: bool = dataclasses.field(default=True)
    group_members_assume_state: bool = dataclasses.field(default=True)


@dataclass(kw_only=True, slots=True)
class DeviceOptions:
    """ZHA device options."""

    enable_identify_on_join: bool = dataclasses.field(default=True)
    consider_unavailable_mains: int = dataclasses.field(
        default=CONF_DEFAULT_CONSIDER_UNAVAILABLE_MAINS
    )
    consider_unavailable_battery: int = dataclasses.field(
        default=CONF_DEFAULT_CONSIDER_UNAVAILABLE_BATTERY
    )
    enable_mains_startup_polling: bool = dataclasses.field(default=True)


@dataclass(kw_only=True, slots=True)
class AlarmControlPanelOptions:
    """ZHA alarm control panel options."""

    master_code: str = dataclasses.field(default="1234")
    failed_tries: int = dataclasses.field(default=3)
    arm_requires_code: bool = dataclasses.field(default=False)


@dataclass(kw_only=True, slots=True)
class CoordinatorConfiguration:
    """ZHA coordinator configuration."""

    path: str
    baudrate: int = dataclasses.field(default=115200)
    flow_control: str = dataclasses.field(default="hardware")
    radio_type: str = dataclasses.field(default="ezsp")


@dataclass(kw_only=True, slots=True)
class QuirksConfiguration:
    """ZHA quirks configuration."""

    enabled: bool = dataclasses.field(default=True)
    custom_quirks_path: str | None = dataclasses.field(default=None)


@dataclass(kw_only=True, slots=True)
class DeviceOverridesConfiguration:
    """ZHA device overrides configuration."""

    type: Platform


@dataclass(kw_only=True, slots=True)
class ZHAConfiguration:
    """ZHA configuration."""

    coordinator_configuration: CoordinatorConfiguration = dataclasses.field(
        default_factory=CoordinatorConfiguration
    )
    quirks_configuration: QuirksConfiguration = dataclasses.field(
        default_factory=QuirksConfiguration
    )
    device_overrides: dict[str, DeviceOverridesConfiguration] = dataclasses.field(
        default_factory=dict
    )
    light_options: LightOptions = dataclasses.field(default_factory=LightOptions)
    device_options: DeviceOptions = dataclasses.field(default_factory=DeviceOptions)
    alarm_control_panel_options: AlarmControlPanelOptions = dataclasses.field(
        default_factory=AlarmControlPanelOptions
    )


@dataclasses.dataclass(kw_only=True, slots=True)
class ZHAData:
    """ZHA data stored in `gateway.data`."""

    config: ZHAConfiguration
    zigpy_config: dict[str, Any] = dataclasses.field(default_factory=dict)
    platforms: collections.defaultdict[Platform, list] = dataclasses.field(
        default_factory=lambda: collections.defaultdict(list)
    )
    gateway: Gateway | None = dataclasses.field(default=None)
    device_trigger_cache: dict[str, tuple[str, dict]] = dataclasses.field(
        default_factory=dict
    )
    allow_polling: bool = dataclasses.field(default=False)
    local_timezone: datetime.tzinfo = dataclasses.field(default=datetime.UTC)


class GlobalUpdater:
    """Global updater for ZHA.

    This class is used to update all listeners at a regular interval. The listeners
    are `Callable` objects that are registered with the `register_update_listener` method.
    """

    _REFRESH_INTERVAL = (30, 45)
    __polling_interval: int

    def __init__(self, gateway: Gateway):
        """Initialize the GlobalUpdater."""
        self._updater_task_handle: asyncio.Task | None = None
        self._update_listeners: list[Callable] = []
        self._gateway: Gateway = gateway

    def start(self):
        """Start the global updater."""
        self._updater_task_handle = self._gateway.async_create_background_task(
            self.update_listeners(),
            name=f"global-updater_{self.__class__.__name__}",
            eager_start=True,
            untracked=True,
        )
        _LOGGER.debug(
            "started global updater with an interval of %s seconds",
            getattr(self, "__polling_interval"),
        )

    def stop(self):
        """Stop the global updater."""
        _LOGGER.debug("stopping global updater")
        if self._updater_task_handle:
            self._updater_task_handle.cancel()
            self._updater_task_handle = None
        _LOGGER.debug("global updater stopped")

    def register_update_listener(self, listener: Callable):
        """Register an update listener."""
        if listener in self._update_listeners:
            _LOGGER.debug(
                "listener already registered with global updater - nothing to register: %s",
                listener,
            )
            return
        self._update_listeners.append(listener)

    def remove_update_listener(self, listener: Callable):
        """Remove an update listener."""
        if listener not in self._update_listeners:
            _LOGGER.debug(
                "listener not registered with global updater - nothing to remove: %s",
                listener,
            )
            return
        self._update_listeners.remove(listener)

    @periodic(_REFRESH_INTERVAL)
    async def update_listeners(self):
        """Update all listeners."""
        _LOGGER.debug("Global updater interval starting")
        if self._gateway.config.allow_polling:
            for listener in self._update_listeners:
                _LOGGER.debug("Global updater running update callback")
                listener()
        else:
            _LOGGER.debug("Global updater interval skipped")
        _LOGGER.debug("Global updater interval finished")


class DeviceAvailabilityChecker:
    """Device availability checker for ZHA."""

    _REFRESH_INTERVAL = (30, 45)
    __polling_interval: int

    def __init__(self, gateway: Gateway):
        """Initialize the DeviceAvailabilityChecker."""
        self._gateway: Gateway = gateway
        self._device_availability_task_handle: asyncio.Task | None = None

    def start(self):
        """Start the device availability checker."""
        self._device_availability_task_handle = (
            self._gateway.async_create_background_task(
                self.check_device_availability(),
                name=f"device-availability-checker_{self.__class__.__name__}",
                eager_start=True,
                untracked=True,
            )
        )
        _LOGGER.debug(
            "started device availability checker with an interval of %s seconds",
            getattr(self, "__polling_interval"),
        )

    def stop(self):
        """Stop the device availability checker."""
        _LOGGER.debug("stopping device availability checker")
        if self._device_availability_task_handle:
            self._device_availability_task_handle.cancel()
            self._device_availability_task_handle = None
        _LOGGER.debug("device availability checker stopped")

    @periodic(_REFRESH_INTERVAL)
    async def check_device_availability(self):
        """Check device availability."""
        _LOGGER.debug("Device availability checker interval starting")
        if self._gateway.config.allow_polling:
            _LOGGER.debug("Checking device availability")
            # 20 because most devices will not make remote calls
            await gather_with_limited_concurrency(
                20,
                *(
                    dev._check_available()
                    for dev in self._gateway.devices.values()
                    if not dev.is_coordinator
                ),
            )
            _LOGGER.debug("Device availability checker interval finished")
        else:
            _LOGGER.debug("Device availability checker interval skipped")



"""Tests for debouncer implementation."""

import asyncio
import logging
from unittest.mock import AsyncMock, Mock

import pytest

from zha.application.gateway import Gateway
from zha.debounce import Debouncer

_LOGGER = logging.getLogger(__name__)


async def test_immediate_works(zha_gateway: Gateway) -> None:
    """Test immediate works."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=AsyncMock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Call when cooldown active setting execute at end to True
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True
    assert debouncer._job.target == debouncer.function

    # Canceling debounce in cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    before_job = debouncer._job

    # Call and let timer run out
    await debouncer.async_call()
    assert len(calls) == 2
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function
    assert debouncer._job == before_job

    # Test calling doesn't execute/cooldown if currently executing.
    await debouncer._execute_lock.acquire()
    await debouncer.async_call()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_immediate_works_with_schedule_call(zha_gateway: Gateway) -> None:
    """Test immediate works with scheduled calls."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=AsyncMock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Call when cooldown active setting execute at end to True
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True
    assert debouncer._job.target == debouncer.function

    # Canceling debounce in cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    before_job = debouncer._job

    # Call and let timer run out
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function
    assert debouncer._job == before_job

    # Test calling doesn't execute/cooldown if currently executing.
    await debouncer._execute_lock.acquire()
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_immediate_works_with_callback_function(zha_gateway: Gateway) -> None:
    """Test immediate works with callback function."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=Mock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    debouncer.async_cancel()


async def test_immediate_works_with_executor_function(zha_gateway: Gateway) -> None:
    """Test immediate works with executor function."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=Mock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    debouncer.async_cancel()


async def test_immediate_works_with_passed_callback_function_raises(
    zha_gateway: Gateway,
) -> None:
    """Test immediate works with a callback function that raises."""
    calls: list[None] = []

    def _append_and_raise() -> None:
        calls.append(None)
        raise RuntimeError("forced_raise")

    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=_append_and_raise,
    )

    # Call when nothing happening
    with pytest.raises(RuntimeError, match="forced_raise"):
        await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Call when cooldown active setting execute at end to True
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True
    assert debouncer._job.target == debouncer.function

    # Canceling debounce in cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    before_job = debouncer._job

    # Call and let timer run out
    with pytest.raises(RuntimeError, match="forced_raise"):
        await debouncer.async_call()
    assert len(calls) == 2
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function
    assert debouncer._job == before_job

    # Test calling doesn't execute/cooldown if currently executing.
    await debouncer._execute_lock.acquire()
    await debouncer.async_call()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_immediate_works_with_passed_coroutine_raises(
    zha_gateway: Gateway,
) -> None:
    """Test immediate works with a coroutine that raises."""
    calls: list[None] = []

    async def _append_and_raise() -> None:
        calls.append(None)
        raise RuntimeError("forced_raise")

    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=_append_and_raise,
    )

    # Call when nothing happening
    with pytest.raises(RuntimeError, match="forced_raise"):
        await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Call when cooldown active setting execute at end to True
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True
    assert debouncer._job.target == debouncer.function

    # Canceling debounce in cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    before_job = debouncer._job

    # Call and let timer run out
    with pytest.raises(RuntimeError, match="forced_raise"):
        await debouncer.async_call()
    assert len(calls) == 2
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function
    assert debouncer._job == before_job

    # Test calling doesn't execute/cooldown if currently executing.
    await debouncer._execute_lock.acquire()
    await debouncer.async_call()
    assert len(calls) == 2
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_not_immediate_works(zha_gateway: Gateway) -> None:
    """Test immediate works."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.8,
        immediate=False,
        function=AsyncMock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    await debouncer.async_call()
    assert len(calls) == 0
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True

    # Call while still on cooldown
    await debouncer.async_call()
    assert len(calls) == 0
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True

    # Canceling while on cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False

    # Call and let timer run out
    await debouncer.async_call()
    assert len(calls) == 0
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Reset debouncer
    debouncer.async_cancel()

    # Test calling doesn't schedule if currently executing.
    await debouncer._execute_lock.acquire()
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_not_immediate_works_schedule_call(zha_gateway: Gateway) -> None:
    """Test immediate works with schedule call."""
    calls: list[None] = []
    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.8,
        immediate=False,
        function=AsyncMock(side_effect=lambda: calls.append(None)),
    )

    # Call when nothing happening
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 0
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True

    # Call while still on cooldown
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 0
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True

    # Canceling while on cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False

    # Call and let timer run out
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 0
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Reset debouncer
    debouncer.async_cancel()

    # Test calling doesn't schedule if currently executing.
    await debouncer._execute_lock.acquire()
    debouncer.async_schedule_call()
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_immediate_works_with_function_swapped(zha_gateway: Gateway) -> None:
    """Test immediate works and we can change out the function."""
    calls: list[None] = []

    one_function = AsyncMock(side_effect=lambda: calls.append(1))
    two_function = AsyncMock(side_effect=lambda: calls.append(2))

    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=True,
        function=one_function,
    )

    # Call when nothing happening
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    # Call when cooldown active setting execute at end to True
    await debouncer.async_call()
    assert len(calls) == 1
    assert debouncer._timer_task is not None
    assert debouncer._execute_at_end_of_timer is True
    assert debouncer._job.target == debouncer.function

    # Canceling debounce in cooldown
    debouncer.async_cancel()
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function

    before_job = debouncer._job
    debouncer.function = two_function

    # Call and let timer run out
    await debouncer.async_call()
    assert len(calls) == 2
    assert calls == [1, 2]
    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 2
    assert calls == [1, 2]
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    assert debouncer._job.target == debouncer.function
    assert debouncer._job != before_job

    # Test calling doesn't execute/cooldown if currently executing.
    await debouncer._execute_lock.acquire()
    await debouncer.async_call()
    assert len(calls) == 2
    assert calls == [1, 2]
    assert debouncer._timer_task is None
    assert debouncer._execute_at_end_of_timer is False
    debouncer._execute_lock.release()
    assert debouncer._job.target == debouncer.function


async def test_shutdown(zha_gateway: Gateway, caplog: pytest.LogCaptureFixture) -> None:
    """Test shutdown."""
    calls: list[None] = []
    future = asyncio.get_running_loop().create_future()

    async def _func() -> None:
        await future
        calls.append(None)

    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.01,
        immediate=False,
        function=_func,
    )

    # Ensure shutdown during a run doesn't create a cooldown timer
    zha_gateway.async_create_task(debouncer.async_call())
    await asyncio.sleep(0.01)
    debouncer.async_shutdown()
    future.set_result(True)
    await zha_gateway.async_block_till_done()
    assert len(calls) == 1
    assert debouncer._timer_task is None

    assert "Debouncer call ignored as shutdown has been requested." not in caplog.text
    await debouncer.async_call()
    assert "Debouncer call ignored as shutdown has been requested." in caplog.text

    assert len(calls) == 1
    assert debouncer._timer_task is None


async def test_background(zha_gateway: Gateway) -> None:
    """Test background tasks are created when background is True."""
    calls: list[None] = []

    async def _func() -> None:
        await asyncio.sleep(0.5)
        calls.append(None)

    debouncer = Debouncer(
        zha_gateway,
        _LOGGER,
        cooldown=0.8,
        immediate=True,
        function=_func,
        background=True,
    )

    await debouncer.async_call()
    assert len(calls) == 1

    debouncer.async_schedule_call()
    assert len(calls) == 1

    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done(wait_background_tasks=False)
    assert len(calls) == 1

    await zha_gateway.async_block_till_done(wait_background_tasks=True)
    assert len(calls) == 2

    await asyncio.sleep(1)
    await zha_gateway.async_block_till_done(wait_background_tasks=False)
    assert len(calls) == 2



"""Helpers for sensor platform."""

from math import ceil, log10


def resolution_to_decimal_precision(
    resolution: float, *, epsilon: float = 2**-23, max_digits: int = 16
) -> int:
    """Calculate the decimal precision for the provided resolution."""
    assert resolution > 0

    threshold = epsilon * max(abs(resolution), 1.0)

    # If the resolution can be closely represented by a fraction n/10^d for some d, we
    # can assume that this is what was intended as the "real" resolution that was then
    # modified during float32 conversion
    for d in range(0, max_digits):
        k = 10**d
        n = round(resolution * k)

        if n == 0:
            continue

        # `abs(resolution - n / k) <= threshold` with less division
        if abs(resolution * k - n) <= threshold * k:
            return d

    # If nothing was found, fall back to the number of decimal places in epsilon
    return ceil(-log10(epsilon))



"""Security cluster handlers module for Zigbee Home Automation."""

from __future__ import annotations

from collections.abc import Callable
import dataclasses
from typing import TYPE_CHECKING, Any, Final

import zigpy.zcl
from zigpy.zcl.clusters.security import (
    IasAce as AceCluster,
    IasWd,
    IasZone,
    Squawk,
    Strobe,
    StrobeLevel,
    WarningType,
)

from zha.exceptions import ZHAException
from zha.zigbee.cluster_handlers import ClusterHandler, ClusterHandlerStatus, registries
from zha.zigbee.cluster_handlers.const import CLUSTER_HANDLER_STATE_CHANGED

if TYPE_CHECKING:
    from zha.zigbee.endpoint import Endpoint

SIGNAL_ARMED_STATE_CHANGED = "zha_armed_state_changed"
SIGNAL_ALARM_TRIGGERED = "zha_armed_triggered"


@dataclasses.dataclass(frozen=True, kw_only=True)
class ClusterHandlerStateChangedEvent:
    """Event to signal that a cluster attribute has been updated."""

    event_type: Final[str] = "cluster_handler_event"
    event: Final[str] = "cluster_handler_state_changed"


@registries.CLUSTER_HANDLER_REGISTRY.register(AceCluster.cluster_id)
class IasAceClusterHandler(ClusterHandler):
    """IAS Ancillary Control Equipment cluster handler."""

    def __init__(self, cluster: zigpy.zcl.Cluster, endpoint: Endpoint) -> None:
        """Initialize IAS Ancillary Control Equipment cluster handler."""
        super().__init__(cluster, endpoint)
        self.command_map: dict[int, Callable[..., Any]] = {
            AceCluster.ServerCommandDefs.arm.id: self.arm,
            AceCluster.ServerCommandDefs.bypass.id: self._bypass,
            AceCluster.ServerCommandDefs.emergency.id: self._emergency,
            AceCluster.ServerCommandDefs.fire.id: self._fire,
            AceCluster.ServerCommandDefs.panic.id: self.panic,
            AceCluster.ServerCommandDefs.get_zone_id_map.id: self._get_zone_id_map,
            AceCluster.ServerCommandDefs.get_zone_info.id: self._get_zone_info,
            AceCluster.ServerCommandDefs.get_panel_status.id: self._send_panel_status_response,
            AceCluster.ServerCommandDefs.get_bypassed_zone_list.id: self._get_bypassed_zone_list,
            AceCluster.ServerCommandDefs.get_zone_status.id: self._get_zone_status,
        }
        self.arm_map: dict[AceCluster.ArmMode, Callable[..., Any]] = {
            AceCluster.ArmMode.Disarm: self._disarm,
            AceCluster.ArmMode.Arm_All_Zones: self._arm_away,
            AceCluster.ArmMode.Arm_Day_Home_Only: self._arm_day,
            AceCluster.ArmMode.Arm_Night_Sleep_Only: self._arm_night,
        }
        self.armed_state: AceCluster.PanelStatus = AceCluster.PanelStatus.Panel_Disarmed
        self.invalid_tries: int = 0

        # These will all be setup by the entity from ZHA configuration
        self.panel_code: str = "1234"
        self.code_required_arm_actions = False
        self.max_invalid_tries: int = 3

        # where do we store this to handle restarts
        self.alarm_status: AceCluster.AlarmStatus = AceCluster.AlarmStatus.No_Alarm

    def cluster_command(self, tsn, command_id, args) -> None:
        """Handle commands received to this cluster."""
        self.debug(
            "received command %s", self._cluster.server_commands[command_id].name
        )
        self.command_map[command_id](*args)

    def arm(self, arm_mode: int, code: str | None, zone_id: int) -> None:
        """Handle the IAS ACE arm command."""
        mode = AceCluster.ArmMode(arm_mode)

        self.emit_zha_event(
            AceCluster.ServerCommandDefs.arm.name,
            {
                "arm_mode": mode.value,
                "arm_mode_description": mode.name,
                "code": code,
                "zone_id": zone_id,
            },
        )

        zigbee_reply = self.arm_map[mode](code)
        self._endpoint.device.gateway.async_create_task(zigbee_reply)

        if self.invalid_tries >= self.max_invalid_tries:
            self.alarm_status = AceCluster.AlarmStatus.Emergency
            self.armed_state = AceCluster.PanelStatus.In_Alarm
            self.emit_zha_event(f"{self.unique_id}_{SIGNAL_ALARM_TRIGGERED}", [])
        else:
            self.emit_zha_event(f"{self.unique_id}_{SIGNAL_ARMED_STATE_CHANGED}", [])
        self._emit_panel_status_changed()

    def _disarm(self, code: str):
        """Test the code and disarm the panel if the code is correct."""
        if (
            code != self.panel_code
            and self.armed_state != AceCluster.PanelStatus.Panel_Disarmed
        ):
            self.debug("Invalid code supplied to IAS ACE")
            self.invalid_tries += 1
            zigbee_reply = self.arm_response(
                AceCluster.ArmNotification.Invalid_Arm_Disarm_Code
            )
        else:
            self.invalid_tries = 0
            if (
                self.armed_state == AceCluster.PanelStatus.Panel_Disarmed
                and self.alarm_status == AceCluster.AlarmStatus.No_Alarm
            ):
                self.debug("IAS ACE already disarmed")
                zigbee_reply = self.arm_response(
                    AceCluster.ArmNotification.Already_Disarmed
                )
            else:
                self.debug("Disarming all IAS ACE zones")
                zigbee_reply = self.arm_response(
                    AceCluster.ArmNotification.All_Zones_Disarmed
                )

            self.armed_state = AceCluster.PanelStatus.Panel_Disarmed
            self.alarm_status = AceCluster.AlarmStatus.No_Alarm
        return zigbee_reply

    def _arm_day(self, code: str) -> None:
        """Arm the panel for day / home zones."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Stay,
            AceCluster.ArmNotification.Only_Day_Home_Zones_Armed,
        )

    def _arm_night(self, code: str) -> None:
        """Arm the panel for night / sleep zones."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Night,
            AceCluster.ArmNotification.Only_Night_Sleep_Zones_Armed,
        )

    def _arm_away(self, code: str) -> None:
        """Arm the panel for away mode."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Away,
            AceCluster.ArmNotification.All_Zones_Armed,
        )

    def _handle_arm(
        self,
        code: str,
        panel_status: AceCluster.PanelStatus,
        armed_type: AceCluster.ArmNotification,
    ) -> None:
        """Arm the panel with the specified statuses."""
        if self.code_required_arm_actions and code != self.panel_code:
            self.debug("Invalid code supplied to IAS ACE")
            zigbee_reply = self.arm_response(
                AceCluster.ArmNotification.Invalid_Arm_Disarm_Code
            )
        else:
            self.debug("Arming all IAS ACE zones")
            self.armed_state = panel_status
            zigbee_reply = self.arm_response(armed_type)
        return zigbee_reply

    def _bypass(self, zone_list, code) -> None:
        """Handle the IAS ACE bypass command."""
        self.emit_zha_event(
            AceCluster.ServerCommandDefs.bypass.name,
            {"zone_list": zone_list, "code": code},
        )

    def _emergency(self) -> None:
        """Handle the IAS ACE emergency command."""
        self._set_alarm(AceCluster.AlarmStatus.Emergency)

    def _fire(self) -> None:
        """Handle the IAS ACE fire command."""
        self._set_alarm(AceCluster.AlarmStatus.Fire)

    def panic(self) -> None:
        """Handle the IAS ACE panic command."""
        self._set_alarm(AceCluster.AlarmStatus.Emergency_Panic)

    def _set_alarm(self, status: AceCluster.AlarmStatus) -> None:
        """Set the specified alarm status."""
        self.alarm_status = status
        self.armed_state = AceCluster.PanelStatus.In_Alarm
        self._emit_panel_status_changed()

    def _get_zone_id_map(self):
        """Handle the IAS ACE zone id map command."""

    def _get_zone_info(self, zone_id):
        """Handle the IAS ACE zone info command."""

    def _send_panel_status_response(self) -> None:
        """Handle the IAS ACE panel status response command."""
        response = self.panel_status_response(
            self.armed_state,
            0x00,
            AceCluster.AudibleNotification.Default_Sound,
            self.alarm_status,
        )
        self._endpoint.device.gateway.async_create_task(response)

    def _emit_panel_status_changed(self) -> None:
        """Handle the IAS ACE panel status changed command."""
        response = self.panel_status_changed(
            self.armed_state,
            0x00,
            AceCluster.AudibleNotification.Default_Sound,
            self.alarm_status,
        )
        self._endpoint.device.gateway.async_create_task(response)
        self.emit(
            CLUSTER_HANDLER_STATE_CHANGED,
            ClusterHandlerStateChangedEvent(),
        )

    def _get_bypassed_zone_list(self):
        """Handle the IAS ACE bypassed zone list command."""

    def _get_zone_status(
        self, starting_zone_id, max_zone_ids, zone_status_mask_flag, zone_status_mask
    ):
        """Handle the IAS ACE zone status command."""


@registries.CLUSTER_HANDLER_ONLY_CLUSTERS.register(IasWd.cluster_id)
@registries.CLUSTER_HANDLER_REGISTRY.register(IasWd.cluster_id)
class IasWdClusterHandler(ClusterHandler):
    """IAS Warning Device cluster handler."""

    @staticmethod
    def set_bit(destination_value, destination_bit, source_value, source_bit):
        """Set the specified bit in the value."""

        if IasWdClusterHandler.get_bit(source_value, source_bit):
            return destination_value | (1 << destination_bit)
        return destination_value

    @staticmethod
    def get_bit(value, bit):
        """Get the specified bit from the value."""
        return (value & (1 << bit)) != 0

    async def issue_squawk(
        self,
        mode=Squawk.SquawkMode.Armed,
        strobe=Strobe.Strobe,
        squawk_level=Squawk.SquawkLevel.High_level_sound,
    ):
        """Issue a squawk command.

        This command uses the WD capabilities to emit a quick audible/visible
        pulse called a "squawk". The squawk command has no effect if the WD
        is currently active (warning in progress).
        """
        value = 0
        value = IasWdClusterHandler.set_bit(value, 0, squawk_level, 0)
        value = IasWdClusterHandler.set_bit(value, 1, squawk_level, 1)

        value = IasWdClusterHandler.set_bit(value, 3, strobe, 0)

        value = IasWdClusterHandler.set_bit(value, 4, mode, 0)
        value = IasWdClusterHandler.set_bit(value, 5, mode, 1)
        value = IasWdClusterHandler.set_bit(value, 6, mode, 2)
        value = IasWdClusterHandler.set_bit(value, 7, mode, 3)

        await self.squawk(value)

    async def issue_start_warning(
        self,
        mode=WarningType.WarningMode.Emergency,
        strobe=Strobe.Strobe,
        siren_level=WarningType.SirenLevel.High_level_sound,
        warning_duration=5,  # seconds
        strobe_duty_cycle=0x00,
        strobe_intensity=StrobeLevel.High_level_strobe,
    ):
        """Issue a start warning command.

        This command starts the WD operation. The WD alerts the surrounding area
        by audible (siren) and visual (strobe) signals.

        strobe_duty_cycle indicates the length of the flash cycle. This provides a means
        of varying the flash duration for different alarm types (e.g., fire, police,
        burglar). Valid range is 0-100 in increments of 10. All other values SHALL
        be rounded to the nearest valid value. Strobe SHALL calculate duty cycle over
        a duration of one second.

        The ON state SHALL precede the OFF state. For example, if Strobe Duty Cycle
        Field specifies “40,” then the strobe SHALL flash ON for 4/10ths of a second
        and then turn OFF for 6/10ths of a second.
        """
        value = 0
        value = IasWdClusterHandler.set_bit(value, 0, siren_level, 0)
        value = IasWdClusterHandler.set_bit(value, 1, siren_level, 1)

        value = IasWdClusterHandler.set_bit(value, 2, strobe, 0)

        value = IasWdClusterHandler.set_bit(value, 4, mode, 0)
        value = IasWdClusterHandler.set_bit(value, 5, mode, 1)
        value = IasWdClusterHandler.set_bit(value, 6, mode, 2)
        value = IasWdClusterHandler.set_bit(value, 7, mode, 3)

        await self.start_warning(
            value, warning_duration, strobe_duty_cycle, strobe_intensity
        )


@registries.CLUSTER_HANDLER_REGISTRY.register(IasZone.cluster_id)
class IASZoneClusterHandler(ClusterHandler):
    """Cluster handler for the IASZone Zigbee cluster."""

    _value_attribute: str = IasZone.AttributeDefs.zone_status.name

    ZCL_INIT_ATTRS = {
        IasZone.AttributeDefs.zone_status.name: False,
        IasZone.AttributeDefs.zone_state.name: True,
        IasZone.AttributeDefs.zone_type.name: True,
    }

    def cluster_command(self, tsn, command_id, args):
        """Handle commands received to this cluster."""
        if command_id == IasZone.ClientCommandDefs.status_change_notification.id:
            zone_status = args[0]
            # update attribute cache with new zone status
            self.cluster.update_attribute(
                IasZone.AttributeDefs.zone_status.id, zone_status
            )
            self.debug("Updated alarm state: %s", zone_status)
        elif command_id == IasZone.ClientCommandDefs.enroll.id:
            self.debug("Enroll requested")
            self._cluster.create_catching_task(
                self.enroll_response(
                    enroll_response_code=IasZone.EnrollResponse.Success, zone_id=0
                )
            )

    async def async_configure(self):
        """Configure IAS device."""
        await self.get_attribute_value(
            IasZone.AttributeDefs.zone_type.name, from_cache=False
        )
        if self._endpoint.device.skip_configuration:
            self.debug("skipping IASZoneClusterHandler configuration")
            return

        self.debug("started IASZoneClusterHandler configuration")

        await self.bind()
        ieee = self.cluster.endpoint.device.application.state.node_info.ieee

        try:
            await self.write_attributes_safe(
                {IasZone.AttributeDefs.cie_addr.name: ieee}
            )
            self.debug(
                "wrote cie_addr: %s to '%s' cluster",
                str(ieee),
                self._cluster.ep_attribute,
            )
        except ZHAException as ex:
            self.debug(
                "Failed to write cie_addr: %s to '%s' cluster: %s",
                str(ieee),
                self._cluster.ep_attribute,
                str(ex),
            )

        self.debug("Sending pro-active IAS enroll response")
        self._cluster.create_catching_task(
            self.enroll_response(
                enroll_response_code=IasZone.EnrollResponse.Success, zone_id=0
            )
        )

        self._status = ClusterHandlerStatus.CONFIGURED
        self.debug("finished IASZoneClusterHandler configuration")

    async def async_update(self) -> None:
        """Retrieve latest state."""
        await self.get_attribute_value(
            IasZone.AttributeDefs.zone_status.name, from_cache=False
        )
